#include "stdafx.h"
#include <math.h>


//---------------------------------------------------------------------------------
const int TOKEN_VAR = -200;				// passed parameter t0, t1, t2 = -200, -199, -198, ... 
	// allow up to 100 variables ? 

const int TOKEN_ADD = -2;			// Binary...
const int TOKEN_SUB = -3;
const int TOKEN_MUL = -4; 
const int TOKEN_DIV = -5;
const int TOKEN_POW = -6;
const int TOKEN_MAX = -7;			// Function...
const int TOKEN_MIN = -8;
const int TOKEN_SIN = -9;			
const int TOKEN_COS = -10;
const int TOKEN_LN  = -11;
const int TOKEN_EXP = -12;
const int TOKEN_SGN = -13;
const int TOKEN_ABS = -14;
const int TOKEN_OP  = -15;			// Other...
const int TOKEN_CP  = -16;
const int TOKEN_COMMA= -17;
const int TOKEN_LINE = -18;			// Line(u,v) from u to v
const int TOKEN_RAMP = -19;			// Ramp(t,u,v,u) from u to v during t, then from v to u
const int TOKEN_PULSE= -20;			// Pulse(t,u,v) u during t then v
const int TOKEN_RMPEX= -21;			// Rmpex(t,u,v,u) from u to v during t, then from v to u
const int TOKEN_RND  = -22;			//
const int TOKEN_SQUARE= -23;			// 
const int TOKEN_HAI	 = -24;

const int TOKEN_NULL = -9999;
const int TOKEN_VALUE = 0; // Token >=0 : Index to double array 

const int FCOUNT = 15;
char * g_apchFunctionNames[FCOUNT]= { 
			"sgn", "abs", "sin", "cos", 
			"ln", "exp", "max", "min", 
			"line", "ramp", "pulse", "rmpex", 
			"rnd",	"square", "hai" };

int g_aiFunctionTokens[FCOUNT]= { 
			TOKEN_SGN,	TOKEN_ABS, TOKEN_SIN, TOKEN_COS, 
			TOKEN_LN,	TOKEN_EXP, TOKEN_MAX, TOKEN_MIN, 
			TOKEN_LINE,	TOKEN_RAMP, TOKEN_PULSE, TOKEN_RMPEX, 
			TOKEN_RND,	TOKEN_SQUARE, TOKEN_HAI };

int g_aiFunctionArgcount[FCOUNT]={ 
			1, 1, 1, 1, 
			1, 1, 2, 2, 
			2, 4, 3, 3, 
			1, 4, 6 };

const int OPCOUNT = 8;		// note includes non operators ( ) 
char g_achOperators[OPCOUNT] = { '+', '-', '*', '/', '^', '(', ')', ',' };
int g_aiOperatorTokens[OPCOUNT] = { TOKEN_ADD, TOKEN_SUB, TOKEN_MUL, TOKEN_DIV, 
									TOKEN_POW, TOKEN_OP, TOKEN_CP, TOKEN_COMMA };

//--- constructor / destructor 
CExpression :: CExpression() { 
	m_aTokens=NULL;
	m_aValues=NULL;
	m_aStack=NULL;
	m_achVarnames=NULL;
	m_apchVars=NULL;
	m_bValid=false;
	}

CExpression :: CExpression( const char * pchExpr, char * achVariables ) { 
	m_aTokens=NULL;
	m_aValues=NULL;
	m_aStack=NULL;
	m_achVarnames=NULL;
	m_apchVars=NULL;
	Parse( pchExpr, achVariables );
	}

CExpression :: ~CExpression() { 
	if( m_achVarnames != NULL ) 
		delete [] m_achVarnames;
	if( m_apchVars != NULL ) 
		delete [] m_apchVars;
	if( m_aTokens != NULL ) 
		delete [] m_aTokens;
	if( m_aValues != NULL ) 
		delete [] m_aValues;
	if( m_aStack != NULL ) 
		delete [] m_aStack;
	}

// Set piToken and pnValue for next token or return false.  pchExpr must be lowercase.
bool CExpression :: NextToken( const char * pchExpr, int * piPos, int * piToken, int iPrevToken, double * pnValue ) { 
	int ilen, isub, i, iFun;
	char ch;

	// skip whitespace
	while( *(pchExpr+(*piPos))==' ' || *(pchExpr+(*piPos))=='\t' ) 
		++(*piPos);

	// Check for empty string
	ilen = strlen(pchExpr+(*piPos));
	if( ilen == 0 ) 
		return false;

	// Check for funcs ( sgn, abs, sin, cos, ln, exp, ...  )
	for( i=0; i<FCOUNT; ++i ) {
		iFun=strlen(g_apchFunctionNames[i]);
		
		if( ilen>=iFun && memicmp(pchExpr+*piPos,g_apchFunctionNames[i],iFun)==0 ) {
			*piToken = g_aiFunctionTokens[i];
			(*piPos) += iFun;
			return true;
			}
		}
			
	// Check for variable names // revers order so ch15 precedes ch1 which is a prefix of it.
	for( i=m_iVars-1; i>=0; --i ) {
		iFun = strlen(m_apchVars[i]);
		if( ilen>=iFun && 0==memcmp( m_apchVars[i], pchExpr+*piPos, iFun ) ) {
			*piToken = TOKEN_VAR+i;
			(*piPos) += iFun;
			return true;
			}
		}

	// Check for operators and other single char tokens
	ch = *(pchExpr+*piPos); 
	for( i=0; i<OPCOUNT; ++i ) {
		if( ch==g_achOperators[i] && ( ch!='-' || IsValue(iPrevToken) || iPrevToken==TOKEN_CP ) ) {
			*piToken = g_aiOperatorTokens[i];
			++(*piPos);
			return true;
			}
		}

	// check for all others
	char ch0; 
	if( ch=='-' || ch=='.' || ( ch>='0' && ch<='9') ) { 
		bool bDecimal=false;
		*piToken=TOKEN_VALUE;
		isub=0; 
		while( strchr("0123456789.", (ch0=*(pchExpr+(*piPos)+isub+1)) )!=NULL 
					&& ( ch0!='.' || ! bDecimal ) ) {
			if( ch0=='.' )
				bDecimal=true;
			++isub;
			}
			
		char * achBuf = new char[isub+2];
		memcpy( achBuf, pchExpr+(*piPos), isub+1 );
		achBuf[isub+1]=0;
		*pnValue = atof(pchExpr+(*piPos));
		*piPos+=isub+1;
		delete [] achBuf;
		return true;
		}
	return false;
	}

bool CExpression :: ParseVariables( char * pchVar ) { 
	//-- count variables 
	int iLen;
	m_iVars=0;
	char * pch = pchVar;
	do { 
		iLen = strlen(pch);
		pch += iLen+1;
		++m_iVars;
		} while( pch[0]!=0 );
	if( m_iVars==0 ) 
		return true;

	//-- allocate space 		
	int iBufLen = pch-pchVar+1;
	if( m_apchVars != NULL ) 
		delete [] m_apchVars;
	if( m_achVarnames != NULL ) 
		delete [] m_achVarnames;
	m_achVarnames = new char[iBufLen];
	m_apchVars = new char * [m_iVars];
	if( m_achVarnames==NULL || m_apchVars==NULL ) 
		return false;
	memcpy( m_achVarnames, pchVar, iBufLen );

	//-- fill in ptrs to vars
	int iVar=0;
	pch = pchVar;
	do { 
		m_apchVars[iVar++]=pch;
		int iLen = strlen(pch);
		pch += iLen+1;
		} while( pch[0]!=0 );
	return true;
	}

bool CExpression :: Parse( const char * pchExpr, char * achVariables ) { 
	m_bValid=false;
	ParseVariables( achVariables );

	if( pchExpr==NULL || strlen(pchExpr)==0 ) 
		pchExpr="1.0";

	if( m_aTokens != NULL ) 
		delete [] m_aTokens;
	if( m_aValues != NULL ) 
		delete [] m_aValues;
	if( m_aStack != NULL ) 
		delete [] m_aStack;
	m_aStack=NULL;
	m_aValues=NULL;
	m_aTokens=NULL;

	//----- count tokens to create array 
	int iTokens=0, iPos=0, iValues=0, iParen=0;
	double nValue;
	int iToken=TOKEN_NULL;
	while( NextToken( pchExpr, &iPos, &iToken, iToken, &nValue ) ) {
		if( iToken==0 ) 
			++iValues;
		if( iToken==TOKEN_OP )
			++iParen;
		if( iToken==TOKEN_CP )
			if( --iParen < 0 ) 
				return false;
		++iTokens;
		}
	if( iParen != 0 )		// check if Open Parens count = closed parens count
		return false;

	//----- save tokens and values to arrays
	m_aRaw=new int[iTokens];
	m_aValues=new double[iValues];
	iPos=0;
	iValues=0;
	iToken=0;
	int iPrevToken=TOKEN_NULL;
	while( NextToken( pchExpr, &iPos, &m_aRaw[iToken], iPrevToken, &nValue ) ) {
		iPrevToken=m_aRaw[iToken];
		if( m_aRaw[iToken]==TOKEN_VALUE ) {
			m_aRaw[iToken]=iValues;
			m_aValues[iValues++]=nValue;
			}
		++iToken;
		}

	//----- reorder tokens into m_aTokens array 
	m_aTokens=new int[iTokens];
	for( iToken=0; iToken<iTokens; ++iToken ) 
		m_aTokens[iToken]=TOKEN_NULL;
	ParseReduce( 0, iTokens-1 );
	delete [] m_aRaw;

	//----- Finish Token Array processing 
	m_iStackSize=0;
	int iStack=0;
	m_iTokens=0;
	for( iToken=0; iToken<iTokens; ++iToken ) {
		//-- pack m_aTokens array 
		if( m_aTokens[iToken]!=TOKEN_NULL ) 
			m_aTokens[m_iTokens++]=m_aTokens[iToken];
		//-- Calculate stacksize and pre allocate 
		if ( IsBinary(m_aTokens[iToken]) ) 
			iStack--;
		else if( IsValue(m_aTokens[iToken]) || m_aTokens[iToken]==TOKEN_COMMA  ) 
			if( (++iStack) > m_iStackSize ) 
				m_iStackSize=iStack;
		// todo - function evals - reduce stack size by Argcount-1 ? 
		}
	m_aStack=new double[m_iStackSize+1];	// +1 for safety on expressions with parse error
	
	return m_bValid=true;
	}

int CExpression :: ParseFind( int iToken1, int iToken2, int i0, int i1 ) { 
	for( int i=i0; i<=i1; ++i ) 
		if( m_aRaw[i]==iToken1 || m_aRaw[i]==iToken2 ) 
			return i;
	return -1;
	}

int CExpression :: ParseFindL( int iToken1, int iToken2, int i0, int i1 ) { 
	for( int i=i1; i>=i0; --i ) 
		if( m_aRaw[i]==iToken1 || m_aRaw[i]==iToken2 ) 
			return i;
	return -1;
	}

void CExpression :: ParseShiftL( int i0, int i1, int iShift ) { 
	for( int i=i0; i<=i1; ++i ) 
		m_aTokens[i-iShift]=m_aTokens[i];
	}

int CExpression :: FindVar( int iLow, int iHigh ) { 
	// return first variable in range specified or -1
	for( int iTkn=0; iTkn<m_iTokens; ++iTkn ) 
		if( m_aTokens[iTkn]>=TOKEN_VAR+iLow && m_aTokens[iTkn]<=TOKEN_VAR+iHigh ) 
			return m_aTokens[iTkn]-TOKEN_VAR;
	return -1; 
	}

void CExpression :: ParseReduce( int i0, int i1 ) { 
	//----- Repeatedly Evaluate innermost parentheses
	int iCP, iOP;
	while( (iCP=ParseFind(TOKEN_CP, TOKEN_CP, i0, i1)) >= 0 )  { 
		iOP = ParseFindL( TOKEN_OP, TOKEN_OP,  i0, iCP-1 );
		//-- reduce expr in (...)
		ParseReduce( iOP+1, iCP-1 );
		//-- remove parentheses
		m_aRaw[iOP]=TOKEN_NULL;
		m_aRaw[iCP]=TOKEN_NULL;
		//-- if parentheses are a function eval, copy the function
		if( iOP>0 && IsFunction(m_aRaw[iOP-1]) ) { 
			m_aTokens[iCP]=m_aRaw[iOP-1];
			m_aRaw[iOP-1]=TOKEN_NULL;
			}
		}

	//-----Evaluate comma delimited lists
	int iOpPos;
	if ( (iOpPos=ParseFindL( TOKEN_COMMA, TOKEN_COMMA, i0, i1))>0 ) { 
		ParseReduce( i0, iOpPos-1 );
		ParseReduce( iOpPos+1, i1 );
		ParseShiftL( iOpPos+1, i1 );
		m_aTokens[i1]=TOKEN_COMMA;
		m_aRaw[iOpPos]=TOKEN_NULL;
		}

	//-----Evaluate Binary :  +,- then *,/ then ^
	else if ( (iOpPos=ParseFindL( TOKEN_ADD, TOKEN_SUB, i0, i1))>0 
			|| (iOpPos=ParseFindL( TOKEN_MUL, TOKEN_DIV, i0, i1))>0 
			|| (iOpPos=ParseFindL( TOKEN_POW, TOKEN_POW, i0, i1))>0 ) { 
		ParseReduce( i0, iOpPos-1 );
		if( iOpPos+1<=i1 ) { 
			ParseReduce( iOpPos+1, i1 );
			ParseShiftL( iOpPos+1, i1 );
			}
		m_aTokens[i1]=m_aRaw[iOpPos];
		m_aRaw[iOpPos]=TOKEN_NULL;
		}

	//----- Evaluate Values 
	else if( IsValue(m_aRaw[i0]) ) {
		m_aTokens[i0]=m_aRaw[i0];
		m_aRaw[i0]=TOKEN_NULL;
		}
	}

double CExpression :: Evaluate( double * aVar, bool * pbError ) { 
	int iStack=0, iFunc;
	double t, f, a;
	bool bError=false;
	
	if( pbError != NULL ) 
		*pbError=false;


	for( int i=0; i<m_iTokens; ++i ) {
		// check for correct number of args for each function type 
		if( IsBinary(m_aTokens[i]) )
			bError = iStack<2;
		else
			for( iFunc=0; iFunc<FCOUNT; ++iFunc ) {
				if( m_aTokens[i]==g_aiFunctionTokens[iFunc] ) { 
					bError = (iStack<g_aiFunctionArgcount[iFunc]);
					break;
					}
				}
		if( bError ) { 
			if( pbError != NULL ) 
				*pbError=true;
			return 0.0;
			}
			
		// evaluate 1 token 
		switch( m_aTokens[i] ) { 
			case TOKEN_MAX :
				m_aStack[iStack-2]=max(m_aStack[iStack-2],m_aStack[iStack-1]);
				iStack--;
				break;
			case TOKEN_MIN :
				m_aStack[iStack-2]=min(m_aStack[iStack-2],m_aStack[iStack-1]);
				iStack--;
				break;
			case TOKEN_ADD :
				m_aStack[iStack-2]+=m_aStack[iStack-1];
				iStack--;
				break;
			case TOKEN_SUB :
				m_aStack[iStack-2]-=m_aStack[iStack-1];
				iStack--;
				break;
			case TOKEN_MUL :
				m_aStack[iStack-2]*=m_aStack[iStack-1];
				iStack--;
				break;
			case TOKEN_DIV :
				if( m_aStack[iStack-1] != 0.0 ) 
					m_aStack[iStack-2]/=m_aStack[iStack-1];
				else if( pbError != NULL ) 
					*pbError=true;
				iStack--;
				break;
			case TOKEN_POW :
				m_aStack[iStack-2]=pow(m_aStack[iStack-2],m_aStack[iStack-1]);
				iStack--;
				break;
		
			case TOKEN_EXP :
				m_aStack[iStack-1]=exp(m_aStack[iStack-1]);
				break;
			case TOKEN_SIN :
				m_aStack[iStack-1]=sin(m_aStack[iStack-1]);
				break;
			case TOKEN_COS :
				m_aStack[iStack-1]=cos(m_aStack[iStack-1]);
				break;
			case TOKEN_LN  :
				m_aStack[iStack-1]=log(m_aStack[iStack-1]);
				break;
			case TOKEN_ABS :
				m_aStack[iStack-1]=fabs(m_aStack[iStack-1]);
				break;
			case TOKEN_SGN :
				m_aStack[iStack-1]=(m_aStack[iStack-1]>0?1.0:(m_aStack[iStack-1]<0?-1.0:0.0));
				break;
			case TOKEN_RND :
				m_aStack[iStack-1] = double(rand()) * m_aStack[iStack-1] / RAND_MAX;
				break;
			case TOKEN_COMMA :
				break;

			// the next 3 func's assume aVar[1] = [0..1] (fraction complete of segment).
			case TOKEN_RAMP :
				t = aVar[1];
				f = m_aStack[iStack-4];
				
				m_aStack[iStack-4] = t<f
						? (1-t/f)*m_aStack[iStack-3]+t/f*m_aStack[iStack-2]
						: (1-t)/(1-f)*m_aStack[iStack-2]+(t-f)/(1-f)*m_aStack[iStack-1] ;
				iStack-=3;
				break;

			case TOKEN_RMPEX :
				t = aVar[1];
				a = m_aStack[iStack-3];
				f = (m_aStack[iStack-1]-m_aStack[iStack-2]) / (exp(a*0.5)-1);
				m_aStack[iStack-3] = t<0.5 
						? (exp(a*t)-1) * f + m_aStack[iStack-2]
						: (exp(a*(1-t))-1) * f + m_aStack[iStack-2];
				iStack-=2;
				break;

			case TOKEN_PULSE :
				t = aVar[1];
				m_aStack[iStack-3] = t<m_aStack[iStack-3]?m_aStack[iStack-2]:m_aStack[iStack-1];
				iStack-=2;
				break;
				
			case TOKEN_SQUARE :
				t = aVar[1];
				m_aStack[iStack-4] =  (t<m_aStack[iStack-4] || t>(1-m_aStack[iStack-3]) ) 
					? m_aStack[iStack-2] : m_aStack[iStack-1];
				t<m_aStack[iStack-3]?m_aStack[iStack-2]:m_aStack[iStack-1];
				iStack-=3;
				break;

			case TOKEN_LINE :
				t = aVar[1];
				m_aStack[iStack-2] = (1-t)*m_aStack[iStack-2]+t*m_aStack[iStack-1] ;
				iStack--;
				break;

			case TOKEN_HAI: {
				double t = m_aStack[iStack-6];
				m_aStack[iStack-6] = (aVar[0]<m_aStack[iStack-4] ? m_aStack[iStack-5] : 
									(aVar[0]<m_aStack[iStack-2] ? m_aStack[iStack-3] : m_aStack[iStack-1] ));
				if( t==0 || t==1.0 ) 
					m_aStack[iStack-6]=0.0;
				else if( aVar[1]<t ) // linear translation ( ramp up ) over initial t time to final value
					m_aStack[iStack-6]*=aVar[1]/t;
				else if( aVar[1]>1.0-t ) // linear translation ( ramp down ) over final t time to 0
					m_aStack[iStack-6]*=(1.0-aVar[1])/t;
				iStack-=5;
				}
				break;

			default : 
				if( iStack<0 || iStack >= m_iStackSize ) { 
					#ifdef _DEBUG
					OutputDebugString("Stack overflow");
					#endif
					*pbError=true;
					return 0.0;
					}
					
				if( m_aTokens[i]>=TOKEN_VAR && m_aTokens[i]<TOKEN_VAR+m_iVars )
					m_aStack[iStack++]=aVar[m_aTokens[i]-TOKEN_VAR];
				else
					m_aStack[iStack++]=m_aValues[m_aTokens[i]];
			}
		}

	if( iStack<1 ) {
		if( pbError != NULL ) 
			*pbError=true;
		return 0.0;
		}
	return m_aStack[0];
	}

bool CExpression :: IsFunction( int iToken ) { 
	for( int i=0; i<FCOUNT; ++i ) 
		if( iToken==g_aiFunctionTokens[i] ) 
			return true;
	return false;
	}
	
bool CExpression :: IsValue( int iToken ) {
	return iToken>=0 || (iToken>=TOKEN_VAR && iToken<TOKEN_VAR+m_iVars) ;
	}
	
bool CExpression :: IsBinary( int iToken ) { 
	return iToken==TOKEN_ADD || iToken==TOKEN_SUB || iToken==TOKEN_MUL 
		|| iToken==TOKEN_DIV || iToken==TOKEN_POW;
	}


#ifdef _DEBUG
void CExpression :: Text( char * achBuffer, int iBufSize ) {
	int i=0, iPos=0;
	while( iPos+20<iBufSize && i<m_iTokens ){
		switch( m_aTokens[i] ) {
			case TOKEN_POW :							// EXPONENT power
				achBuffer[iPos++]='^';
				break;
			case TOKEN_ADD :								// +
				achBuffer[iPos++]='+';
				break;
			case TOKEN_COMMA :								// ,
				achBuffer[iPos++]=',';
				break;
			case TOKEN_SUB :								// -
				achBuffer[iPos++]='-';
				break;
			case TOKEN_MUL :								// *
				achBuffer[iPos++]='*';
				break;
			case TOKEN_DIV :								// /
				achBuffer[iPos++]='/';
				break;
			case TOKEN_MAX  :
				memcpy(achBuffer+iPos,"max",3);
				iPos+=3;
				break;
			case TOKEN_MIN  :
				memcpy(achBuffer+iPos,"min",3);
				iPos+=3;
				break;
			case TOKEN_LINE  :
				memcpy(achBuffer+iPos,"line",4);
				iPos+=4;
				break;
			case TOKEN_PULSE  :
				memcpy(achBuffer+iPos,"pulse",5);
				iPos+=5;
				break;
			case TOKEN_RAMP  :
				memcpy(achBuffer+iPos,"ramp",4);
				iPos+=4;
				break;
			case TOKEN_RMPEX  :
				memcpy(achBuffer+iPos,"rmpex",6);
				iPos+=6;
				break;
			case TOKEN_SIN  :
				memcpy(achBuffer+iPos,"sin",3);
				iPos+=3;
				break;
			case TOKEN_COS  :
				memcpy(achBuffer+iPos,"cos",3);
				iPos+=3;
				break;
			case TOKEN_LN   :
				memcpy(achBuffer+iPos,"ln",2);
				iPos+=2;
				break;
			case TOKEN_EXP  :
				memcpy(achBuffer+iPos,"exp",3);
				iPos+=3;
				break;
			case TOKEN_ABS  :
				memcpy(achBuffer+iPos,"abs",3);
				iPos+=3;
				break;
			case TOKEN_SGN  :
				memcpy(achBuffer+iPos,"sgn",3);
				iPos+=3;
				break;
			case TOKEN_RND  :
				memcpy(achBuffer+iPos,"rnd",3);
				iPos+=3;
				break;
			case TOKEN_NULL :
				memcpy(achBuffer+iPos,"null",4);
				iPos+=4;
				break;
			default : 
				if( m_aTokens[i]>=TOKEN_VAR && m_aTokens[i]<TOKEN_VAR+m_iVars ) {
					sprintf(achBuffer+iPos, "%s", m_apchVars[m_aTokens[i]-TOKEN_VAR]);
					iPos=strlen(achBuffer);
					}
				else if( m_aTokens[i] >= 0 ) { 
					sprintf(achBuffer+iPos, "%f", m_aValues[m_aTokens[i]]);
					iPos=strlen(achBuffer);
					}
				else 
					achBuffer[iPos++]='?';	
				break;
			}
		++i;
		achBuffer[iPos++]=' ';
		}
	achBuffer[iPos++]=0;
	}
#endif

